# file: svm.py
# author: Bob Muller
# date: February 21, 2013
# revised: November 1, 2013
# revised: March, 2016
#
# This program implements the Simple Virtual Machine (SVM).
#
# Symbolic constants naming the 16 SVM operations.
#
(LOD, STO, MOV, HLT)      = (0, 1, 2, 3)
(ADD, SUB, MUL, DIV)      = (4, 5, 6, 7)
(CMP, LI,  JSR, R)        = (8, 9, 10, 11)
(JMP, BLT, BEQ, BGT)      = (12, 13, 14, 15)

def instrToString(instr):
   front = ["LOD", "STO", "MOV", "HLT", "ADD", "SUB", "MUL", "DIV"]
   back  = ["CMP", "LI", "JSR", "R", "JMP", "BLT", "BEQ", "BGT"]
   all = front + back
   return all[instr]

# Symbolic constants naming the general purpose registers.
#
(R0, R1, R2, R3) = (0, 1, 2, 3)
ZERO = 4

def regToString(r):
    return ["R0", "R1", "R2", "R3", "ZERO"][r]

# A sample SVM program. Let M = data1[0], N = data1[1].
# Then pgm1 computes R0 = M * N.
#
data1 = [4, 5]
text1 = [
    (MOV, [R0, ZERO]),
    (LI,  [R3, 1]),
    (LOD, [R1, 0, ZERO]),
    (LOD, [R2, 1, ZERO]),
    (CMP, [R2, ZERO]),
    (BEQ, [3]),
    (ADD, [R0, R0, R1]),
    (SUB, [R2, R2, R3]),
    (JMP, [-5]),
    (HLT, [])
    ]
image1 = data1 + text1

# image1 is typically executed on svm with the call: svm(len(data1), image1).

DEBUG = True
STEP = False

# Register and RAM put and get helper functions.
#
def registerPut(value, dest, registers):
    [r0, r1, r2, r3] = registers
    if dest == 0:
        return (value, r1, r2, r3)
    elif dest == 1:
        return (r0, value, r2, r3)
    elif dest == 2:
        return (r0, r1, value, r3)
    else:
        return (r0, r1, r2, value)

def registerGet(r, registers):
   if r == ZERO:
      return 0
   else:
      return registers[r]

def ramPut(value, address, ram):
   return ram[:address] + [value] + ram[address+1:]

def ramGet(addr, ram):
   return ram[addr]

# printState
#
def printState(pc, psw, ra, registers):
    print 'PC = ' + str(pc)
    print 'PSW = ' + str(psw)
    print 'RA = ' + str(ra)
    print 'R0 = ' + str(registers[0])
    print 'R1 = ' + str(registers[1])
    print 'R2 = ' + str(registers[2])
    print 'R3 = ' + str(registers[3])

# An implementation of the Simple Virtual Machine. A call svm(pc, image)
# executes the SVM program starting with instruction image[pc].
#
def svm(pc, image):

    # The CPU Instruction cycle.
    #
    def cycle(pc, psw, ra, registers, image):

        # Fetch the next instruction from image.  Instructions are
        # a represented as a pair (opcode, [opnd1, opnd2, ...]).
        #
        (opcode, operands) = image[pc]
        newPC = pc + 1

        # Set DEBUG above to True to get diagnostic information.
        if DEBUG:
            printState(pc, psw, ra, registers)
            print '\ndbg: instr = ' + str(image[pc])

            if STEP:
                _ = raw_input()

        # Now dispatch on the opcode.
        #
        if opcode == LOD:      # LOD  dst, offset, src
            dst = operands[0]
            offset = operands[1]
            src = operands[2]
            address = offset + registerGet(src, registers)
            value = ramGet(address, image)
            newRegisters = registerPut(value, dst, registers)
            cycle(newPC, psw, ra, newRegisters, image)
        
        elif opcode == STO:     # STO   src, offset, dst
            src = operands[0]
            offset = operands[1]
            dst = operands[2]
            value = registerGet(src, registers)
            address = offset + registerGet(dst, registers)
            newRAM = ramPut(value, address, image)
            cycle(newPC, psw, ra, registers, newRAM)

        elif opcode == MOV:        # MOV  dst, src
            dst = operands[0]
            src = operands[1]
            value = registerGet(src, registers)
            newRegisters = registerPut(value, dst, registers)
            cycle(newPC, psw, ra, newRegisters, image)

        elif opcode == ADD:        # ADD  dst, src1, src2
            dst = operands[0]
            src1 = operands[1]
            src2 = operands[2]
            v1 = registerGet(src1, registers)
            v2 = registerGet(src2, registers)
            newRegisters = registerPut(v1 + v2, dst, registers)
            cycle(newPC, psw, ra, newRegisters, image)

        elif opcode == SUB:        # SUB  dst, src1, src2
            dst = operands[0]
            src1 = operands[1]
            src2 = operands[2]
            v1 = registerGet(src1, registers)
            v2 = registerGet(src2, registers)
            newRegisters = registerPut(v1 - v2, dst, registers)
            cycle(newPC, psw, ra, newRegisters, image)

        elif opcode == MUL:        # MUL  dst, src1, src2
            dst = operands[0]
            src1 = operands[1]
            src2 = operands[2]
            v1 = registerGet(src1, registers)
            v2 = registerGet(src2, registers)
            newRegisters = registerPut(v1 * v2, dst, registers)
            cycle(newPC, psw, ra, newRegisters, image)

        elif opcode == DIV:        # DIV  dst, src1, src2
            dst = operands[0]
            src1 = operands[1]
            src2 = operands[2]
            v1 = registerGet(src1, registers)
            v2 = registerGet(src2, registers)
            newRegisters = registerPut(v1 / v2, dst, registers)
            cycle(newPC, psw, ra, newRegisters, image)

        elif opcode == CMP:     # CMP   src1, src2
            src1 = operands[0]
            src2 = operands[1]
            v1 = registerGet(src1, registers)
            v2 = registerGet(src2, registers)
            cycle(newPC, v1 - v2, ra, registers, image)

        elif opcode == LI:     # LI     dst, number
            dst = operands[0]
            number = operands[1]
            newRegisters = registerPut(number, dst, registers)
            cycle(newPC, psw, ra, newRegisters, image)

        elif opcode == JSR:     # JSR     displacement
            disp = operands[0]
            cycle(newPC + disp, psw, newPC, registers, image)

        elif opcode == R:       # R
            cycle(ra, psw, ra, registers, image)

        elif opcode == BLT:     # BLT  displacement
            displacement = operands[0]
            if psw < 0:
                cycle(newPC + displacement, psw, ra, registers, image)
            else:
                cycle(newPC, psw, ra, registers, image)

        elif opcode == BEQ:     # BEQ   displacement
            displacement = operands[0]
            if psw == 0:
                cycle(newPC + displacement, psw, ra, registers, image)
            else:
                cycle(newPC, psw, ra, registers, image)

        elif opcode == BGT:     # BGT    displacement
            displacement = operands[0]
            if psw > 0:
                cycle(newPC + displacement, psw, ra, registers, image)
            else:
                cycle(newPC, psw, ra, registers, image)

        elif opcode == JMP:     # JMP    displacement
            displacement = operands[0]
            cycle(newPC + displacement, psw, ra, registers, image)

        elif opcode == HLT:
            print "SVM Halt.\n"
            printState(pc, psw, ra, registers)
            return
        else:
            print 'SVM: unknown opcode ' + str(opcode) + ', blue screen.'

    initialPSW = 0
    initialRA = 0
    initialRegisters = (0, 0, 0, 0)

    # Start the instruction cycle.
    #
    cycle(pc, initialPSW, initialRA, initialRegisters, image)
